"""Handles requests and data un-entanglement to a truly retarded web service."""

from __future__ import annotations

import asyncio
import binascii
from dataclasses import dataclass
from http import HTTPStatus
import logging
from typing import TypeVar
from urllib.parse import urlencode

from aiohttp import ClientError, ClientResponse, ClientResponseError, ClientSession
import async_timeout

from .constants import (
    API_URL,
    CLIENT_TIMEOUT,
    CLIENT_VERSION_NUMBER,
    DEFAULT_COUNTRY,
    ERROR_RESPONSES,
    NO_ACCESS_RESPONSES,
    NO_CONTENT_RESPONSES,
    PHONE_PREFIXES,
)
from .exceptions import (
    AuthenticationBlockedError,
    AuthenticationError,
    NoAccessError,
    NoContentError,
    NotActivatedError,
    NotFoundError,
    PolitikontrollerConnectionError,
    PolitikontrollerError,
    PolitikontrollerTimeoutError,
)
from .models import (
    Account,
    AuthenticationResponse,
    AuthStatus,
    PoliceControlResponse,
    PoliceControlsResponse,
    PoliceGPSControlsResponse,
    PolitiKontrollerResponse,
    UserMap,
)
from .models.api import (
    APIEndpoint,
    EndpointRegistry,
    PoliceControlTypeEnum,
    PolitiKontrollerRequest,
)
from .utils import (
    aes_decrypt,
    aes_encrypt,
    map_response_data,
    merge_duplicate_controls,
)

ResponseT = TypeVar(
    "ResponseT",
    bound=PolitiKontrollerResponse | dict[str, any],
)

_LOGGER = logging.getLogger(__name__)


@dataclass
class Client:
    user: Account | None = None
    session: ClientSession | None = None
    request_timeout: int = CLIENT_TIMEOUT

    _close_session: bool = False

    @classmethod
    def initialize(cls, username: str, password: str, session: ClientSession | None = None) -> Client:
        return cls(Account(username=username, password=password), session=session)

    @classmethod
    async def login(cls, username: str, password: str, session: ClientSession | None = None) -> Client:
        c = cls(session=session)
        await c.authenticate_user(username, password)
        return c

    @property
    def request_header(self) -> dict[str, str]:
        """Generate a header for HTTP requests to the server."""
        return {
            "User-Agent": f"PK_{CLIENT_VERSION_NUMBER}",
        }

    async def api_request(
        self,
        endpoint: APIEndpoint | str,
        params: dict | None = None,
        cast_to: type[ResponseT] | None = None,
        is_list=False,
    ) -> ResponseT | list[ResponseT] | str:
        if params is None:
            params = {}
        if isinstance(endpoint, str):
            endpoint = APIEndpoint.from_str(endpoint)

        # Build request
        request_cls = EndpointRegistry.get_request_class(endpoint)
        if endpoint.requires_auth():
            if self.user is None:
                raise AuthenticationError("Trying to access authenticated API without authentication")
            params["account"] = self.user.to_dict()
        params["p"] = endpoint
        request = request_cls.from_dict(params)

        data = await self.do_external_api_request(request)
        data_parts = data.split("|")
        _LOGGER.debug("Got response: %s", data)

        if data in ERROR_RESPONSES:
            msg = "Unknown error received from Politikontroller.no"
            raise PolitikontrollerError(msg)
        if data in NO_ACCESS_RESPONSES:
            raise NoAccessError
        if data in NO_CONTENT_RESPONSES or len(data) == 0:
            raise NoContentError
        if data_parts[0] in NO_CONTENT_RESPONSES:
            raise NoContentError

        # Attempt to cast the response data to desired model
        if cast_to is not None:
            return cast_to.from_response_data(data, multiple=is_list)

        # Return the raw response (str)
        return data

    def _ensure_session(self):
        if self.session is None or self.session.closed:
            self.session = ClientSession()
            _LOGGER.debug("New session created.")
            self._close_session = True

    @staticmethod
    async def _request_check_status(response: ClientResponse):
        if response.status == HTTPStatus.NOT_FOUND:
            raise NotFoundError("Resource not found")
        if response.status == HTTPStatus.BAD_REQUEST:
            raise PolitikontrollerError("Bad request syntax or unsupported method")
        if response.status == HTTPStatus.FORBIDDEN:
            raise AuthenticationError("Authorization failed")
        if not HTTPStatus(response.status).is_success:
            raise ClientError(response)

    async def do_external_api_request(
        self,
        request: PolitiKontrollerRequest,
        **kwargs,
    ) -> str:
        headers = kwargs.get("headers")
        headers = self.request_header if headers is None else dict(headers)

        payload = request.get_query_params()
        _LOGGER.debug("Doing API request with params: %s", payload)
        url = f"{API_URL}/app.php?{aes_encrypt(urlencode(payload))}"
        headers = {
            "user-agent": f"PK_{CLIENT_VERSION_NUMBER}",
            **headers,
        }

        self._ensure_session()

        try:
            async with async_timeout.timeout(self.request_timeout):
                response = await self.session.get(
                    url,
                    **kwargs,
                    headers=headers,
                    raise_for_status=self._request_check_status,
                )
                enc_data = await response.text("utf-8")
                _LOGGER.debug("Response: %s", enc_data)
                try:
                    data = aes_decrypt(enc_data)
                except (binascii.Error, ValueError):
                    data = enc_data.strip()

                return data

        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to Politikontroller.no"
            raise PolitikontrollerTimeoutError(msg) from exception
        except (
            ClientError,
            ClientResponseError,
        ) as exception:
            raise PolitikontrollerConnectionError(
                f"Error occurred while communicating with Politikontroller.no: {exception}",
            ) from exception
        finally:
            if self._close_session:
                await self.session.close()
                self._close_session = False

    def set_user(self, user: Account):
        self.user = user

    async def authenticate_user(self, username: str, password: str) -> Account:
        """Authenticate user."""
        auth_user = Account(username=username, password=password)
        params = {
            "lang": auth_user.country.lower(),
            "account": auth_user.to_dict(),
        }

        result = await self.api_request(APIEndpoint.LOGIN, params, cast_to=AuthenticationResponse)
        if result.auth_status == AuthStatus.LOGIN_ERROR:
            msg = f"Authentication failed: {result.auth_status}"
            raise AuthenticationError(msg)
        if result.auth_status == AuthStatus.SPERRET:
            raise AuthenticationBlockedError
        if result.auth_status == AuthStatus.NOT_ACTIVATED:
            raise NotActivatedError

        account_dict = {
            **auth_user.to_dict(),
            **result.to_dict(),
            "username": auth_user.username,
        }

        account = Account.from_dict({str(k): str(v) for k, v in account_dict.items()})
        self.set_user(account)
        return account

    async def check(self):
        """Server health check."""
        return await self.api_request(APIEndpoint.CHECK)

    async def get_settings(self):
        """Get settings."""
        try:
            return await self.api_request(APIEndpoint.SETTINGS)
        except NoContentError:
            return {}

    async def get_control(self, cid: int) -> PoliceControlResponse:
        """Get details for a single control."""
        return await self.api_request(
            APIEndpoint.SPEED_CONTROL,
            {
                "kontroll_id": cid,
            },
            cast_to=PoliceControlResponse,
        )

    async def get_controls(
        self,
        lat: float,
        lng: float,
        merge_duplicates: bool = True,
    ) -> list[PoliceControlsResponse]:
        """Get all active controls."""
        try:
            controls = await self.api_request(
                APIEndpoint.SPEED_CONTROLS,
                {
                    "lat": lat,
                    "lon": lng,
                },
                cast_to=PoliceControlsResponse,
                is_list=True,
            )
        except NoContentError:
            return []

        if merge_duplicates:
            return merge_duplicate_controls(controls)
        return controls

    async def get_controls_in_radius(
        self,
        lat: float,
        lng: float,
        radius: int,
        speed: int = 100,
        merge_duplicates: bool = True,
        **kwargs,
    ) -> list[PoliceGPSControlsResponse]:
        """Get all active controls within a radius."""
        params = {
            "vr": radius,
            "speed": speed,
            "lat": lat,
            "lon": lng,
            **kwargs,
        }
        try:
            controls = await self.api_request(
                APIEndpoint.GPS_CONTROLS,
                params,
                cast_to=PoliceGPSControlsResponse,
                is_list=True,
            )
        except NoContentError:
            return []

        if merge_duplicates:
            return merge_duplicate_controls(controls)
        return controls

    async def get_controls_from_lists(
        self,
        controls: list[PoliceGPSControlsResponse | PoliceControlsResponse],
    ) -> list[PoliceControlResponse]:  # pragma: no cover
        """Get details for a list of controls."""
        return [await self.get_control(i.id) for i in controls]

    @staticmethod
    def get_control_types() -> list[PoliceControlTypeEnum]:
        """Get all control types."""
        return [PoliceControlTypeEnum.from_str(i) for i in list(PoliceControlTypeEnum)]

    async def get_maps(self) -> list[UserMap]:
        """Get all user maps."""
        try:
            return await self.api_request(APIEndpoint.GET_MY_MAPS, cast_to=UserMap, is_list=True)
        except NoContentError:
            return []

    async def exchange_points(self):  # pragma: no cover
        """Exchange points."""
        result = await self.api_request(APIEndpoint.EXCHANGE)
        return map_response_data(
            result,
            [
                "status",
                "message",
            ],
        )

    async def account_register(
        self,
        phone_number: int,
        password: str,
        name: str,
        country: str | None = None,
    ):  # pragma: no cover
        if country is None:
            country = DEFAULT_COUNTRY
        country_code = PHONE_PREFIXES.get(country)

        params = {
            "telefon": phone_number,
            "passord": password,
            "cc": country_code,
            "navn": name,
            "lang": country,
        }
        result = await self.api_request(APIEndpoint.REGISTER, params)
        return map_response_data(
            result,
            [
                "status",
                "message",
            ],
        )

    async def account_auth(self, auth_code: str, uid: int):  # pragma: no cover
        """Activate account by auth code."""
        params = {
            "auth_kode": auth_code,
            "uid": uid,
        }
        result = await self.api_request(APIEndpoint.AUTH_APP, params)
        return map_response_data(
            result,
            [
                "status",
                "message",
            ],
        )

    async def account_auth_sms(self):  # pragma: no cover
        """Activate account by sms."""
        result = await self.api_request(APIEndpoint.AUTH_SMS)
        return map_response_data(
            result,
            [
                "status",
                "message",
            ],
        )

    async def account_send_sms(self):  # pragma: no cover
        """Send activation sms."""
        result = await self.api_request(APIEndpoint.SEND_SMS)
        return map_response_data(
            result,
            [
                "status",
                "message",
            ],
        )



from __future__ import annotations

import base64
from datetime import datetime, time as dt_time
from logging import getLogger
from math import atan2, cos, sin, sqrt
import random
import re
import string
import time
from typing import TYPE_CHECKING, TypeVar

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from geojson.mapping import to_mapping

from .constants import (
    CRYPTO_K1,
    CRYPTO_K2,
    DEFAULT_MAX_DISTANCE,
)

if TYPE_CHECKING:
    from .models.api import PoliceControl, PoliceControlPoint

    PC = TypeVar("PC", bound=PoliceControl)

_LOGGER = getLogger(__name__)
JUNK_CHARS = "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x10\x0f"


def get_random_string(length: int, letters: str | None = None) -> str:
    """Generate a random string of given length using given letters."""
    if letters is None:
        letters = string.ascii_uppercase + string.digits
    return "".join(random.choice(letters) for _ in range(length))


def get_unix_timestamp():
    """Get current unix timestamp with 10 seconds added."""
    return int(time.time()) + 10


def aes_encrypt(input_str: str):
    """Encrypts a string using AES encryption with given key and initialization vector.
    Returns base64-encoded result.
    """
    key = base64.b64decode(CRYPTO_K2)
    iv = base64.b64decode(CRYPTO_K1)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    length = 16 - (len(input_str) % 16)
    input_b = bytes(input_str, "utf-8") + bytes([length]) * length
    input_padded = pad(input_b, AES.block_size)
    return base64.b64encode(cipher.encrypt(input_padded)).decode()


def aes_decrypt(enc_base64: str):
    """Decrypts AES encrypted data using a given key and initialization vector."""
    enc_data = base64.b64decode(enc_base64)
    key = base64.b64decode(CRYPTO_K2)
    iv = base64.b64decode(CRYPTO_K1)
    decipher = AES.new(key, AES.MODE_CBC, iv)
    ciphertext_padded = decipher.decrypt(enc_data)
    return unpad(ciphertext_padded, AES.block_size).decode().strip(JUNK_CHARS)


def map_response_data(
    data: str, map_keys: list[str | None], multiple=False
) -> list[dict[str, str]] | dict[str, str]:
    """Convert a cvs-like string into dictionaries."""

    def row_to_dict(row) -> dict[str, str]:
        r = dict(zip(map_keys, row.split("|")))
        return {k: r[k] for k in r if isinstance(k, str)}

    if multiple:
        return list(map(row_to_dict, list(data.split("#"))))

    return row_to_dict(data)


def unmap_response_data(
    data: list[dict[str, any]] | dict[str, any], map_keys: list[str | None]
) -> str:  # pragma: no cover
    """Convert dictionaries into a cvs-like string."""

    def dict_to_row(d: dict[str, any]) -> str:
        row = [str(d.get(k, "")) if k is not None else "" for k in map_keys]
        return "|".join(row)

    if isinstance(data, list):
        return "#".join(map(dict_to_row, data))

    return dict_to_row(data)


def parse_datetime_like(v: str) -> int | None:
    """Parse datetime like string to unix timestamp."""
    if isinstance(v, int) and v != 0:
        return v
    if len(v) == 0 or (v.isnumeric() and int(v) == 0):
        return None
    return int(parse_time_format(v))


def parse_time_format(text: str) -> int | str:
    """Parse time format to unix timestamp."""
    today = datetime.now().astimezone()
    try:
        # Match "%d.%m - %H:%M" this way due to failure on leap days using strptime.
        if m := re.match(r"(\d{2})\.(\d{2}) - (\d{2}):(\d{2})", text):
            return int(
                datetime.fromisoformat(f"{today.year}-{m[2]}-{m[1]}T{m[3]}:{m[4]}:00")
                .astimezone()
                .timestamp()
            )
    except ValueError:  # pragma: no cover
        pass

    try:
        return int(
            datetime.combine(
                today,
                dt_time.fromisoformat(text),
            )
            .astimezone()
            .timestamp()
        )
    except ValueError:  # pragma: no cover
        pass

    try:  # pragma: no cover
        text = re.sub(r"(\d{2}:\d{2})(?: \(\d+ ganger\))?", "\\1", text)
        return int(
            datetime.strptime(text, "%H:%M")
            .astimezone()
            .replace(
                year=today.year,
                month=today.month,
                day=today.day,
            )
            .timestamp()
        )
    except ValueError:  # pragma: no cover
        pass
    return text  # pragma: no cover


def calculate_distance(point1: PoliceControlPoint, point2: PoliceControlPoint) -> float:
    """Calculate distance between two points."""
    earth_radius = 6373.0  # Approximate radius of earth in km
    lat1, lon1 = point1.coordinates(as_radian=True)
    lat2, lon2 = point2.coordinates(as_radian=True)

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    return earth_radius * c


def average_points(p1: PoliceControlPoint, p2: PoliceControlPoint) -> tuple[float, float]:
    """Average two points."""
    return (p1.lat + p2.lat) / 2, (p1.lng + p2.lng) / 2


def merge_duplicate_controls(controls: list[PC], max_distance: float | None = None) -> list[PC]:
    """Merge duplicate controls."""
    if max_distance is None:
        max_distance = DEFAULT_MAX_DISTANCE
    merged_controls = []
    skip_indices = set()

    for i, control1 in enumerate(controls):
        if i in skip_indices:
            continue

        for j, control2 in enumerate(controls):
            if i == j or j in skip_indices:
                continue

            distance = calculate_distance(control1.point, control2.point)
            if distance <= max_distance and control1.type == control2.type:
                control1 = control1.merge_with(control2)  # noqa: PLW2901

                skip_indices.add(j)

        merged_controls.append(control1)

    return merged_controls


def to_geo_json(controls: list[PC]):
    return {
        "type": "FeatureCollection",
        "features": [to_mapping(c) for c in controls],
    }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.types import Discriminator

from politikontroller_py.constants import (
    DEFAULT_COUNTRY,
    PHONE_NUMBER_LENGTH,
    PHONE_PREFIXES,
)
from politikontroller_py.models.common import (
    BaseModel,
    PolitiKontrollerResponse,
    StrEnum,
)
from politikontroller_py.utils import map_response_data

if TYPE_CHECKING:
    from politikontroller_py.models.common import T


def deserialize_yesno(value: str):
    return not (value == "NO" or value[:3] == "NO_")


def serialize_yesno(value: bool, suffix: str | None = None):
    if suffix is None:
        return "YES" if value else "NO"
    return suffix if value else f"NO_{suffix}"


def deserialize_bool(value: str):
    return value == "true"


def serialize_bool(value: bool):
    return "true" if value else "false"


class AuthStatus(StrEnum):
    APP_ERR = "APP_ERR"
    LOGIN_OK = "LOGIN_OK"
    LOGIN_ERROR = "LOGIN_ERROR"
    SPERRET = "SPERRET"
    NOT_ACTIVATED = "NOT_ACTIVATED"
    SKIP_AUTHENTICATION = "SKIP_AUTHENTICATION"


@dataclass(kw_only=True)
class AuthenticationResponse(PolitiKontrollerResponse):
    auth_status: AuthStatus

    class Config(BaseConfig):
        discriminator = Discriminator(
            field="auth_status",
            include_subtypes=True,
        )

    attr_map = [
        "auth_status",
    ]

    @classmethod
    def from_response_data(cls: T, cvs: str, multiple=False) -> T | list[T]:
        data = map_response_data(cvs, cls.attr_map, multiple)
        auth_status = AuthStatus.from_str(data.get("auth_status"))
        subtypes = {
            AuthStatus.SPERRET: AuthenticationResponseBlocked,
            AuthStatus.LOGIN_OK: AuthenticationResponseOK,
            AuthStatus.LOGIN_ERROR: AuthenticationResponseError,
            AuthStatus.NOT_ACTIVATED: AuthenticationResponseNotActivated,
        }
        sub_class = subtypes.get(auth_status, cls)
        data = map_response_data(cvs, sub_class.attr_map, multiple)
        return sub_class.from_dict(data)


@dataclass
class AuthenticationResponseOK(AuthenticationResponse):
    auth_status: AuthStatus = AuthStatus.LOGIN_OK
    premium_key: str = "NO"
    user_level: int | None = None
    phone_prefix: int | None = None
    status: str | None = None
    uid: int | None = None
    nickname: str | None = None
    saphne: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_yesno,
            serialize=lambda v: serialize_yesno(v, suffix="SAPHE"),
        ),
    )
    show_regnr: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_yesno,
            serialize=lambda v: serialize_yesno(v, suffix="REGNR"),
        ),
    )
    premium_price: int | None = None
    enable_points: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_yesno,
            serialize=serialize_yesno,
        ),
    )
    enable_calls: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_yesno,
            serialize=serialize_yesno,
        ),
    )
    needs_gps: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_bool,
            serialize=serialize_bool,
        ),
    )
    gps_radius: int | None = None
    push_notification: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_bool,
            serialize=serialize_bool,
        ),
    )
    sms_notification: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_bool,
            serialize=serialize_bool,
        ),
    )
    points: int | None = None
    exchange_code: bool | None = field(
        default=None,
        metadata=field_options(
            deserialize=deserialize_bool,
            serialize=serialize_bool,
        ),
    )

    attr_map = [
        "auth_status",  # LOGIN_OK
        "premium_key",  # NO
        "user_level",  # 0
        "phone_prefix",  # 47
        "status",  # SKIP_AUTHENTICATION
        "uid",  # 1000
        None,  # 0
        "nickname",
        "saphne",  # NO_SAPHE
        "show_regnr",  # NO_REGNR
        "premium_price",  # 29
        "enable_points",  # NO
        "enable_calls",  # NO
        None,  # +47400008
        "gps_radius",  # 30
        "push_notification",  # true
        "sms_notification",  # false
        "points",  # 62
        "exchange_code",  # false
    ]


@dataclass(kw_only=True)
class AuthenticationResponseBlocked(AuthenticationResponse):
    auth_status: AuthStatus = AuthStatus.SPERRET
    block_reason: str

    attr_map = [
        "auth_status",  # SPERRET
        "block_reason",  # ...
    ]


@dataclass(kw_only=True)
class AuthenticationResponseError(AuthenticationResponse):
    auth_status: AuthStatus = AuthStatus.LOGIN_ERROR
    error: str

    attr_map = [
        "auth_status",  # LOGIN_ERROR
        "error",  # WRONG_USERNAME_OR_PASSWORD
    ]


@dataclass(kw_only=True)
class AuthenticationResponseNotActivated(AuthenticationResponse):
    auth_status: AuthStatus = AuthStatus.NOT_ACTIVATED
    uid: int

    attr_map = [
        "auth_status",  # NOT_ACTIVATED
        "uid",  # 1000
    ]


@dataclass
class AccountBase(BaseModel):
    username: str
    password: str | None = None
    country: str = DEFAULT_COUNTRY

    @classmethod
    def __pre_deserialize__(cls: type[T], d: T) -> T:
        d["username"] = d.get("username", "").replace(" ", "")
        return d

    @property
    def phone_number(self):
        return int(self.username[2:]) if len(self.username) > PHONE_NUMBER_LENGTH else int(self.username)

    @property
    def phone_prefix(self):
        return (
            int(self.username[:2])
            if len(self.username) > PHONE_NUMBER_LENGTH
            else PHONE_PREFIXES.get(self.country.lower())
        )


@dataclass(kw_only=True)
class Account(AccountBase):
    uid: int | None = None
    auth_status: AuthStatus | None = None
    status: str | None = None



CRYPTO_KEY_SIZE = 128
CRYPTO_K1 = "aGFua19lcl9ob21vX3Bvbw=="  # "hank_er_homo_poo"
CRYPTO_K2 = "ZGlsZG9zZXJ2aWNlX3N1eA=="  # "dildoservice_sux"
CLIENT_VERSION_NUMBER = "9.1.0"
CLIENT_OS = "Android"
CLIENT_TIMEOUT = 30
API_URL = "http://app.politikontroller.no"

NO_CONTROLS = "INGEN_KONTROLLER"
INGEN = "INGEN"
INGEN_PAAMELDTE_STEDER = "INGEN_PAAMELDTE_STEDER"
USER_NOT_AUTHORIZED = "USER_NOT_AUTHORIZED"
USER_NOT_AUTHORIZED_NOPREM = "USER_NOT_AUTHORIZED_NOPREM"
INVALID_AUTH = "INVALID_AUTH"
ERR = "ERR"

NO_ACCESS_RESPONSES = [
    USER_NOT_AUTHORIZED,
    USER_NOT_AUTHORIZED_NOPREM,
    INVALID_AUTH,
]
NO_CONTENT_RESPONSES = [
    NO_CONTROLS,
    INGEN,
    INGEN_PAAMELDTE_STEDER,
]
ERROR_RESPONSES = [
    ERR,
]

PHONE_PREFIXES = {
    "no": 47,
    "se": 46,
    "dk": 45,
}
PHONE_NUMBER_LENGTH = 8
DEFAULT_COUNTRY = "no"
DEFAULT_MAX_DISTANCE = 1.5

DESCRIPTION_TRUNCATE_LENGTH = 27
DESCRIPTION_TRUNCATE_SUFFIX = ".."



from .account import Account, AuthenticationResponse, AuthStatus
from .api import (
    PoliceControl,
    PoliceControlResponse,
    PoliceControlsResponse,
    PoliceControlType,
    PoliceGPSControlsResponse,
    UserMap,
)
from .common import PolitiKontrollerResponse

__all__ = [
    "Account",
    "AuthenticationResponse",
    "AuthStatus",
    "PoliceControl",
    "PoliceControlResponse",
    "PoliceControlsResponse",
    "PoliceControlType",
    "PoliceGPSControlsResponse",
    "PolitiKontrollerResponse",
    "UserMap",
]



from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime
from math import radians
import re
from typing import TYPE_CHECKING, ClassVar, Literal, TypeVar

from mashumaro import field_options
from mashumaro.config import BaseConfig

from politikontroller_py.constants import (
    CLIENT_OS,
    CLIENT_VERSION_NUMBER,
    DESCRIPTION_TRUNCATE_LENGTH,
    DESCRIPTION_TRUNCATE_SUFFIX,
)
from politikontroller_py.models.account import Account  # noqa: TCH001
from politikontroller_py.models.common import (
    BaseModel,
    PolitiKontrollerResponse,
    StrEnum,
)
from politikontroller_py.utils import (
    average_points,
    get_random_string,
    get_unix_timestamp,
    parse_datetime_like,
)

if TYPE_CHECKING:
    from politikontroller_py.models.common import T

PC = TypeVar("PC", bound="PoliceControl")


# noinspection SpellCheckingInspection
class PoliceControlTypeEnum(StrEnum):
    SPEED_TRAP = "Fartskontroll"
    BEHAVIOUR = "Belte/mobil"
    TECHNICAL = "Teknisk"
    TRAFFIC_INFO = "Trafikk info"
    TRAFFIC_MESSAGE = "Trafikkmelding"
    OBSERVATION = "Observasjon"
    CUSTOMS = "Toll/grense"
    WEIGHT = "Vektkontroll"
    UNKNOWN = "Ukjent"
    CIVIL_POLICE = "Sivilpoliti"
    MC_CONTROL = "Mopedkontroll"
    BOAT_PATROL = "Politibåten"


class ExchangeStatus(StrEnum):
    EXCHANGE_OK = "EXCHANGE_OK"


class PoliceControlType(BaseModel):
    type: PoliceControlTypeEnum
    slug: str

    # attr_map = [
    #     "slug",
    #     "name",
    #     "id",
    #     None,
    # ]
    #
    # @classmethod
    # def __pre_deserialize__(cls: type[T], d: T) -> T:
    #     # Remove ".png"
    #     d["slug"] = d.get("slug", "")[:-4]
    #     return d


@dataclass
class PoliceControlPoint(BaseModel):
    lat: float
    lng: float
    type: str = "Point"

    def coordinates(self, as_radian=False):
        if as_radian:
            return radians(self.lat), radians(self.lng)
        return self.lat, self.lng

    @property
    def __geo_interface__(self):
        return {
            "type": self.type,
            "coordinates": self.coordinates(),
        }


@dataclass
class PoliceControl(PolitiKontrollerResponse):
    id: int
    county: str
    municipality: str
    description: str
    type: PoliceControlTypeEnum
    lat: float
    lng: float
    point: PoliceControlPoint
    timestamp: datetime | None = field(
        default=None,
        metadata=field_options(
            deserialize=datetime.fromtimestamp,
            serialize=lambda v: int(datetime.timestamp(v)),
        ),
    )
    _merged_with: list[PoliceControl] = field(init=False, default_factory=list)

    @classmethod
    def __pre_deserialize__(cls: type[T], d: T) -> T:
        for k in ["timestamp", "last_seen"]:
            if k in d:
                d[k] = parse_datetime_like(d.get(k, ""))
        d["point"] = {"lat": d["lat"], "lng": d["lng"]}
        return d

    @property
    def duplicates(self) -> list[PoliceControl]:
        return self._merged_with

    @property
    def description_truncated(self):
        trunc_len = DESCRIPTION_TRUNCATE_LENGTH - len(DESCRIPTION_TRUNCATE_SUFFIX)
        return (
            f"{self.description[:trunc_len]}{DESCRIPTION_TRUNCATE_SUFFIX}"
            if len(self.description) > DESCRIPTION_TRUNCATE_LENGTH
            else self.description
        )

    @property
    def title(self):
        return f"{self.type}: {self.description_truncated}"

    @property
    def _geometry(self):
        return PoliceControlPoint(self.lat, self.lng)

    @property
    def __geo_interface__(self):
        return {
            "type": "Feature",
            "geometry": self._geometry.__geo_interface__,
            "properties": {
                "title": self.title,
                "description": self.description,
                "type": self.type,
            },
        }

    # noinspection PyAttributeOutsideInit
    def add_merged(self: PC, other: PC):
        self._merged_with.append(other)
        self.timestamp = max(other.timestamp, self.timestamp)
        if hasattr(self, "last_seen"):
            self.last_seen = max(other.last_seen, self.last_seen)
        if hasattr(self, "confirmed"):  # pragma: no cover
            self.confirmed = max(other.confirmed, self.confirmed)
        if hasattr(self, "speed_limit"):  # pragma: no cover
            self.speed_limit = min(other.speed_limit, self.speed_limit)

    def merge_with(self: PC, other: PC) -> PC:
        lat, lng = average_points(self.point, other.point)
        data = self.to_dict()
        data.update(
            {
                "lat": lat,
                "lng": lng,
            }
        )
        control = type(self).from_dict(data)
        control.add_merged(other)
        control.add_merged(self)
        return control


# noinspection SpellCheckingInspection
@dataclass(kw_only=True)
class PoliceControlResponse(PoliceControl):
    speed_limit: int | None = None
    last_seen: datetime | None = field(
        default=None,
        metadata=field_options(
            deserialize=datetime.fromtimestamp,
            serialize=lambda v: int(datetime.timestamp(v)),
        ),
    )
    confirmed: int = 0

    attr_map = [
        "id",  # 59777                     59790
        "county",  # Møre og Romsdal           Innlandet
        "municipality",  # Kristiansund              Elverum
        "type",  # Belte/mobil               Fartskontroll
        "timestamp",  # 05.12 - 10:41             05.12 - 13:38
        "description",  # Statens veivesen Rensvik  busslomme strandbygda
        "lat",  # 63.1033706005419          60.8831426463528
        "lng",  # 7.82150858039696          11.512297578156
        None,  # ?                         ?
        None,  # ?                         ?
        None,  # kristiansund.png          elverum.png
        None,  # more_og_romsdal.png       innlandet.png
        "speed_limit",  # 50                        0
        None,  # 1                         1
        "last_seen",  # 10:41                     13:38
        None,  # 0                         0          0
        None,  # 2                         2          2
        "confirmed",  # 2                         1          5
    ]


@dataclass(kw_only=True)
class PoliceGPSControlsResponse(PoliceControl):
    attr_map = [
        "id",
        "county",
        "municipality",
        "type",
        "timestamp",
        "description",
        "lat",
        "lng",
        None,
        None,
        None,
        None,
    ]


@dataclass(kw_only=True)
class PoliceControlsResponse(PoliceControl):
    last_seen: datetime | None = field(
        default=None,
        metadata=field_options(
            deserialize=datetime.fromtimestamp,
            serialize=lambda v: int(datetime.timestamp(v)),
        ),
    )

    attr_map = [
        "id",
        "county",
        "municipality",
        "type",
        None,
        "description",
        "lat",
        "lng",
        None,
        None,
        None,
        None,
        "timestamp",
        None,
        None,
        "last_seen",
    ]


@dataclass
class UserMap(PolitiKontrollerResponse):
    id: int
    title: str
    country: str

    attr_map = [
        "id",
        None,
        "title",
        "country",
    ]


@dataclass
class ExchangePointsResponse(PolitiKontrollerResponse):
    status: ExchangeStatus
    message: str


TR = TypeVar("TR", bound="PolitiKontrollerRequest")


class EndpointRegistry:
    _registry: dict[APIEndpoint, type[TR]] = {}

    @classmethod
    def register(cls, endpoint: APIEndpoint) -> callable:
        def decorator(request_class: type[TR]) -> type[TR]:
            request_class.p = endpoint.value
            cls._registry[endpoint] = request_class
            return request_class

        return decorator

    @classmethod
    def get_request_class(cls, endpoint: APIEndpoint) -> type[TR]:
        if endpoint not in cls._registry:  # pragma: no cover
            return PolitiKontrollerRequest
        return cls._registry[endpoint]


# noinspection SpellCheckingInspection
class APIEndpoint(StrEnum):
    AUTH_APP = "auth_app"
    AUTH_SMS = "auth_sms"
    CHANGE_PASSWORD = "endre_passord"
    CHANGE_PASSWORD_CODE = "endrings_kode"
    CHECK = "check"
    CONTROL_TYPES = "kontrolltyper"
    EXCHANGE = "veksle"
    GET_MY_MAPS = "hent_mine_kart"
    GPS_CONTROLS = "gps_kontroller"
    LOGIN = "l"
    REGISTER = "r"
    SEND_SMS = "send_sms"
    SETTINGS = "instillinger"
    SPEED_CONTROL = "hki"
    SPEED_CONTROLS = "hk"

    def requires_auth(self) -> bool:
        no_auth_methods = [
            APIEndpoint.LOGIN,
            APIEndpoint.REGISTER,
            APIEndpoint.CHANGE_PASSWORD,
            APIEndpoint.CHANGE_PASSWORD_CODE,
        ]
        return self not in no_auth_methods


API_ENDPOINTS = Literal[
    APIEndpoint.AUTH_APP,
    APIEndpoint.AUTH_SMS,
    APIEndpoint.CHANGE_PASSWORD,
    APIEndpoint.CHANGE_PASSWORD_CODE,
    APIEndpoint.CHECK,
    APIEndpoint.CONTROL_TYPES,
    APIEndpoint.EXCHANGE,
    APIEndpoint.GET_MY_MAPS,
    APIEndpoint.GPS_CONTROLS,
    APIEndpoint.LOGIN,
    APIEndpoint.REGISTER,
    APIEndpoint.SEND_SMS,
    APIEndpoint.SETTINGS,
    APIEndpoint.SPEED_CONTROL,
    APIEndpoint.SPEED_CONTROLS,
]


@dataclass
class PolitiKontrollerRequestBase(BaseModel):
    class Config(BaseConfig):
        omit_none = True
        allow_deserialization_not_by_alias = True
        serialize_by_alias = True


@dataclass
class PolitiKontrollerRequest(PolitiKontrollerRequestBase):
    p: APIEndpoint
    bac: str = field(default_factory=lambda: get_random_string(10))
    z: str = field(default_factory=lambda: str(get_unix_timestamp()))
    version: str = CLIENT_VERSION_NUMBER
    os: str = CLIENT_OS
    tt: str = field(default_factory=lambda: get_random_string(5))

    BASE_FIELDS: ClassVar[list[str]] = ["bac", "z", "version", "os", "tt"]

    def __post_serialize__(self, d: dict[str, any]) -> dict[str, any]:
        """Replace special characters in the base fields with hyphens."""
        for f in self.BASE_FIELDS:
            if f in d:
                d[f] = re.sub(r"[|#\\\"]", "-", str(d[f]))
        return d

    def get_query_params(self) -> dict[str, str]:
        """Get query parameters in specific order.

        1. bac, z, version, os
        2. All other fields from to_dict()
        3. tt at the end
        """
        all_params = self.to_dict()
        query = OrderedDict()

        # Add first 4 params in order
        for param in self.BASE_FIELDS[:-1]:
            if param in all_params:
                query[param] = all_params[param]

        # Add all remaining params except sanitized fields
        for key, value in all_params.items():
            if key not in self.BASE_FIELDS:
                query[key] = value

        query["tt"] = all_params["tt"]

        return dict(query)


@dataclass
class PolitikontrollerAuthenticatedRequest(PolitiKontrollerRequestBase):
    retning: str = field(init=False)
    telefon: str = field(init=False)
    passord: str = field(init=False)
    account: Account = field(metadata=field_options(serialize="omit"))

    def __post_init__(self):
        self.retning = self.account.phone_prefix
        self.telefon = self.account.phone_number
        self.passord = self.account.password


@EndpointRegistry.register(APIEndpoint.LOGIN)
@dataclass(kw_only=True)
class PolitiKontrollerLoginRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    lang: str


@EndpointRegistry.register(APIEndpoint.SPEED_CONTROLS)
@dataclass(kw_only=True)
class PolitiKontrollerGetControlsRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    lat: float
    lon: float


@EndpointRegistry.register(APIEndpoint.GPS_CONTROLS)
@dataclass(kw_only=True)
class PolitiKontrollerGetControlsInRadiusRequest(
    PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest
):
    vr: int
    lat: float
    lon: float
    speed: int | None = None


@EndpointRegistry.register(APIEndpoint.SPEED_CONTROL)
@dataclass(kw_only=True)
class PolitiKontrollerGetControlRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    kontroll_id: int


@EndpointRegistry.register(APIEndpoint.SETTINGS)
@dataclass
class PolitiKontrollerSettingsRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    pass


@EndpointRegistry.register(APIEndpoint.GET_MY_MAPS)
@dataclass
class PolitiKontrollerGetMapsRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    pass


@dataclass
class PolitiKontrollerGetControlTypesRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    pass


@EndpointRegistry.register(APIEndpoint.CHECK)
@dataclass
class PolitiKontrollerCheckRequest(PolitiKontrollerRequest, PolitikontrollerAuthenticatedRequest):
    pass


@EndpointRegistry.register(APIEndpoint.REGISTER)
@dataclass(kw_only=True)
class PolitiKontrollerRegisterRequest(PolitiKontrollerRequest):
    telefon: str
    passord: str
    cc: str
    navn: str
    lang: str


@EndpointRegistry.register(APIEndpoint.AUTH_APP)
@dataclass(kw_only=True)
class PolitiKontrollerAuthAppRequest(PolitiKontrollerRequest):
    auth_kode: str
    uid: int



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, TypeVar

from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin

from politikontroller_py.utils import map_response_data, unmap_response_data

T = TypeVar("T", bound="DataClassORJSONMixin")


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def from_str(cls, value: str) -> StrEnum:
        return cls(value)

    def __eq__(self, other: str) -> bool:
        return super().__eq__(other)

    def __hash__(self):
        return hash(self.value)


@dataclass
class BaseModel(DataClassORJSONMixin):
    class Config(BaseConfig):
        omit_none = True
        allow_deserialization_not_by_alias = True
        serialize_by_alias = True


@dataclass
class PolitiKontrollerResponse(BaseModel):
    attr_map: ClassVar[list[str]]

    @classmethod
    def from_response_data(cls: T, cvs: str, multiple=False) -> T | list[T]:
        """Convert a cvs-like string into instance(s) of `cls`."""
        data = map_response_data(cvs, cls.attr_map, multiple)
        if multiple:
            return [cls.from_dict(d) for d in data]
        return cls.from_dict(data)

    def to_response_data(self) -> str:  # pragma: no cover
        """Convert a serialized version of `self` into a cvs-like string."""
        return unmap_response_data(self.to_dict(), self.attr_map)



"""CLI tool."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import anyio
import asyncclick as click
from tabulate import tabulate

from politikontroller_py import Client
from politikontroller_py.exceptions import AuthenticationError

if TYPE_CHECKING:
    from asyncclick.core import Context

TABULATE_DEFAULTS = {
    "tablefmt": "rounded_grid",
}


def tabulate_model(data: list[dict], keys: list[str]) -> list[list[str]]:
    result = [keys]
    for item in data:
        row = [item.get(key) for key in keys]
        result.append(row)
    return result


@click.group()
@click.option(
    "--username",
    "-u",
    envvar="POLITIKONTROLLER_USERNAME",
    type=str,
    required=True,
    help="Username (i.e. phone number)",
)
@click.password_option(
    "--password",
    "-p",
    envvar="POLITIKONTROLLER_PASSWORD",
    type=str,
    required=True,
    confirmation_prompt=False,
    help="Password",
)
@click.option("--debug", is_flag=True, help="Set logging level to DEBUG")
@click.pass_context
async def cli(ctx: Context, username: str, password: str, debug: bool):
    """Connect to politikontroller.no and fetch data in a simple way.

    Username and password can be defined using env vars.

    POLITIKONTROLLER_USERNAME
    POLITIKONTROLLER_PASSWORD
    """
    configure_logging(debug)

    ctx.obj = client = Client()

    try:
        user = await client.authenticate_user(username, password)
    except AuthenticationError as err:
        raise click.BadParameter(str(err), param_hint="--username, --password") from err
    click.echo(user)


@cli.command("check", short_help="server health check.")
@click.pass_obj
async def check(obj: Client):
    res = await obj.check()
    click.echo(res)


@cli.command("get-control-types", short_help="get a list of control types.")
@click.pass_obj
async def get_control_types(obj):
    types = await obj.get_control_types()
    click.echo(tabulate([t.to_dict() for t in types], **TABULATE_DEFAULTS))


@cli.command("get-controls", short_help="get a list of all active controls.")
@click.option("--lat", required=True, help="Your position (latitude)")
@click.option("--lng", required=True, help="Your position (longitude)")
@click.pass_obj
async def get_controls(obj, lat: float, lng: float):
    controls = await obj.get_controls(lat, lng)
    click.echo(tabulate([d.to_dict() for d in controls], **TABULATE_DEFAULTS))


@cli.command("get-controls-radius", short_help="get all active controls inside a radius.")
@click.option("--lat", type=float, required=True, help="Radius center (latitude)")
@click.option("--lng", type=float, required=True, help="Radius center (longitude)")
@click.option("--radius", type=int, required=True, metavar="km", help="Radius size in kilometers")
@click.option("--speed", type=int, required=False, metavar="km/h", help="Speed, unknown what this does")
@click.pass_obj
async def get_controls_in_radius(obj: Client, lat: float, lng: float, radius: int, speed: int):
    _controls = await obj.get_controls_in_radius(lat, lng, radius, speed)
    controls = await obj.get_controls_from_lists(_controls)

    lists = tabulate_model(
        [c.to_dict() for c in controls],
        [
            "id",
            "timestamp",
            "type",
            "municipality",
            "description",
            "point",
        ],
    )
    click.echo(tabulate(lists, headers="firstrow", **TABULATE_DEFAULTS))


@cli.command("get-control", short_help="get details on a control.")
@click.argument("control_id", type=int, required=True)
@click.pass_obj
async def get_control(obj: Client, control_id: int):
    control = await obj.get_control(control_id)
    click.echo(
        tabulate(
            tabulate_model(
                [control.to_dict()],
                [
                    "id",
                    "county",
                    "municipality",
                    "description",
                    "type",
                    "lat",
                    "lng",
                    "timestamp",
                ],
            ),
            headers="firstrow",
            **TABULATE_DEFAULTS,
        )
    )


@cli.command("get-maps", short_help="get own maps.")
@click.pass_obj
async def get_maps(obj: Client):
    maps = await obj.get_maps()
    click.echo(tabulate([m.to_dict() for m in maps], **TABULATE_DEFAULTS))


@cli.command("get-settings", short_help="get own settings.")
@click.pass_obj
async def get_settings(obj: Client):
    settings = await obj.get_settings()
    click.echo(tabulate(settings, **TABULATE_DEFAULTS))


@cli.command("exchange-points", short_help="exchange points (?)")
@click.pass_obj
async def exchange_points(obj: Client):
    res = await obj.exchange_points()
    click.echo(tabulate(res, **TABULATE_DEFAULTS))


@cli.command("account-send-sms", short_help="send activation sms")
@click.pass_obj
async def account_send_sms(obj: Client):
    res = await obj.account_send_sms()
    click.echo(res)


@cli.command("account-auth-sms", short_help="activate account by sms")
@click.pass_obj
async def account_auth_sms(obj: Client):
    res = await obj.account_auth_sms()
    click.echo(res)


@cli.command("account-auth", short_help="activate account")
@click.argument("code", type=str, required=True)
@click.argument("uid", type=int, required=True)
@click.pass_obj
async def account_auth(obj: Client, code: str, uid: int):
    res = await obj.account_auth(auth_code=code, uid=uid)
    click.echo(res)


@cli.command("account-register", short_help="register new account")
@click.argument("phone", type=int, required=True)
@click.argument("password", type=str, required=True)
@click.argument("name", type=str, required=True)
@click.argument("country", type=str, required=True)
@click.pass_obj
async def account_register(
    obj: Client,
    phone: int,
    password: str,
    name: str,
    country: str,
):
    res = await obj.account_register(
        phone_number=phone,
        password=password,
        name=name,
        country=country,
    )
    click.echo(res)


def configure_logging(debug: bool = False):
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)


async def amain(args: list[str] | None = None):
    await cli.main(args)


def main():
    # noinspection PyTypeChecker
    anyio.run(amain)


if __name__ == "__main__":
    main()
